/* Code developed for DvCon 2010 
 By Ben Cohen, Srinivasan Venkataramanan, Ajeetha Kumari, and Lisa Piper
 "SystemVerilog Assertions Handbook, 2nd edition"ISBN  878-0-9705394-8-7
  www.systemverilog.us  ben@systemverilog.us
  www.cvcblr.com, info@cvcblr.com

 All code provided in this presentation is distributed
 with *ABSOLUTELY NO SUPPORT* and *NO WARRANTY* from the authors.  
*/
module cache_controller_tb();
  timeunit 1ns;     timeprecision 100ps;  
  //user interface
  logic data_vld, busy;  
  logic rd=0, wr=0;
  logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr;                     
  logic [MEM_DATA_WIDTH-1:0]         rd_data; 
  logic  [MEM_DATA_WIDTH-1:0]         wr_data;  
  // main memory interface 
  logic mem_rd, mem_wr; 
  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] mem_addr; 
  logic [MEM_DATA_WIDTH-1:0]         mem_wr_data;  
  logic  [MEM_DATA_WIDTH-1:0]         mem_rd_data; 
  // Signals needed by checker                        
  logic cache_exists;  // exists at specified address
  logic [MEM_DATA_WIDTH-1:0]         cache_data;
  
  // clock and resets 
  logic clk=1'b1; 
  logic rst_n=1'b0;

  logic [2**MAIN_MEM_ADDRESS_WIDTH-1:0] [MEM_DATA_WIDTH-1 : 0] main_mem; // main memory
  logic [MEM_DATA_WIDTH-1 : 0] test_mem_data;
  assign test_mem_data = main_mem[addr];
  int counter = 0;

   // if data written to main_mem, then data read from main memory should not change
  // NEED to add an assertion here
  property p_wr_rd_memory;
    int v_addr, v_data;
      first_match(($rose(wr), v_addr=addr, v_data=wr_data) ##1 
      !(wr && addr==v_addr)[*1:$] ##1 rd && addr==v_addr) |->
                                   main_mem[addr] ==v_data;
  endproperty : p_wr_rd_memory
  ap_wr_rd_memory : assert property(@ (posedge clk) p_wr_rd_memory);

  cp_wr_rd_memory : cover property(@ (posedge clk) p_wr_rd_memory);

  sequence q_wr_rd_memory;
    int v_addr, v_data;
	 first_match(($rose(wr), v_addr=addr, v_data=wr_data) ##1 
      !(wr && addr==v_addr)[*1:$] ##1 rd && addr==v_addr) ##0 
                                   main_mem[addr] ==v_data;
  endsequence : q_wr_rd_memory
  cq_wr_rd_memory : cover property(@ (posedge clk) q_wr_rd_memory);

  
  always @ (posedge clk) begin 
	if (mem_rd) mem_rd_data<= main_mem[mem_addr];
    if (mem_wr) main_mem[mem_addr] <= mem_wr_data;
  end 
  

  cache_controller cache_controller_1(.*);
  initial forever #5 clk=!clk;
  initial begin
	for (int i=0; i<1024; i=i+1)
	  main_mem[i]=i;
	repeat (3) @ (posedge clk);
	rst_n <= 1'b1;
  end

  initial begin : test
	wait(rst_n);
    #100;     @(negedge clk);
	repeat (4) begin : begin_4wr 
      wr =1; rd = 0;
	  a0 : assert (std::randomize(addr, wr_data));
      addr = addr & 10'b0000001111;
      repeat(5) begin @ (negedge clk) rd=0;  wr=0; end;
	end : begin_4wr

    repeat (1050) begin : r1
      a1 : assert (std::randomize(rd, wr, addr, wr_data));
      addr = addr & 10'b0000001111;
//   	    if (!(counter >30 && counter < 35)) begin : no_errors  // forced errors
      if (1'b1) begin : no_errors		  
	    if (busy) begin rd=0;  wr=0; end
	    else if (rd && wr) wr =0;
		end : no_errors
        @(negedge clk);
	    counter = counter+1;
    end : r1	
     rd=0;  wr=0;
	
    #800 $finish;
  end  : test  
    
	
  
endmodule : cache_controller_tb
























