module cache_mem(
                 input  logic wr_cache,  // instructs cache to write a line 
                 input  logic invalidate, // instructs cache to invalidate a line 
                 input  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr,  // address to determine exists 
                 input  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] wr_addr,  // address to write into cache 
                 input  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] addr2invalidate, // address to invalidate
                 input  logic [MEM_DATA_WIDTH-1:0]         cache_wr_data, // data to write to cache
				 input  logic                              check_cache, // check the cache for exists 
                 output logic [MEM_DATA_WIDTH-1:0]         cache_data,  // data read from cache  
                 output logic  cache_exists,               // cache exists at rd_addr
                 // clock and resets 
                 input  logic clk, reset_n);
timeunit 1ns;   timeprecision 100ps;

  logic [(CACHE_SIZE-1):0] [MEM_DATA_WIDTH-1:0]  cache_mem_data; // cache data storage 
  logic [(CACHE_SIZE-1):0] [MAIN_MEM_ADDRESS_WIDTH-1 : 0] cache_mem_addr; // cache address storage  
  logic [CACHE_SIZE-1:0]    cache_entry; // determines if entry in cache is valid 
  logic [2**MAIN_MEM_ADDRESS_WIDTH-1:0]  cache_valid; // quich search of exists 
 
  initial begin : init
    if (!reset_n)
      for (int i=0; i<= 2**MAIN_MEM_ADDRESS_WIDTH-1; i++) begin : loop1
        cache_valid[i] = 1'b0; // reset fast valid memory 
      end : loop1
      for (int i=0; i <= CACHE_SIZE -1; i++) begin : loop2
          cache_entry[i] <= 1'b0;   // reset the entry definitions 
      end : loop2 
    end : init
  // 
  assign cache_exists =cache_valid[addr];
  // assign cache_data = cache_mem_data[cache_mem_addr[MAIN_MEM_ADDRESS_WIDTH-1:0]];
  

  always @ (posedge clk) begin : cache_mem_handling
    automatic logic success, found_existing_entry, found_empty_line;
//    logic success, found_existing_entry, found_empty_line;	
    automatic int   i, j, found_index;
    success =0;
    found_existing_entry =0;
    found_index =0;
    found_empty_line =0;
    // ****invalidate 
    if (invalidate) begin : if1
      cache_valid[addr2invalidate]<= 1'b0;
      for (int j=0; j<= CACHE_SIZE-1; j=j+1) begin : for1
        //$display("infinite loop here ???  L49");
        if(cache_entry[j]==1 && cache_mem_addr[j] == addr2invalidate) begin : if2 // found empty line
          found_existing_entry= 1'b1;
          found_index= j;
          success=1;
		  // $display("At %t L52 wr_cache found_index=%d", $time, found_index);
          if (found_existing_entry) cache_entry[j] <=1'b0;  //  empty line		  
		  break;
        end : if2 
      end : for1
      ap_nothing2invalidate: assert(found_existing_entry) else $error("nothing to invalidate");
      ap_nothing2invalidate_concurrent: assert property (found_existing_entry) else $error("nothing to invalidate");
    end : if1
    // #2 // Temp fix wr_cache && invalidate simulteneous issues  
    // **** Write to cache
    if (wr_cache && !invalidate) begin : write2cache_noInvalidate
	  found_existing_entry = 1'b0; success =1'b0;
      // Check if address exists and find the index for that address (the j).
      for (int j=0; j<= CACHE_SIZE-1; j=j+1) begin : for0
        // $display("infinite loop here ???  L65");             
        if(cache_entry[j]==1 && cache_mem_addr[j]==wr_addr) begin : if1// found line
          found_existing_entry = 1'b1;
          found_index = j;
          success =1;
		  // $display("L68 wr_cache existing no invalidate, found_index=%d", found_index);
		  break;
          // $display("found existing entry in wr L70");
        end : if1 
	  end : for0

	end : write2cache_noInvalidate   
	if (wr_cache) begin : wr_cache_updates 
      if (found_existing_entry || found_empty_line) begin : if_found_entry// write to the cache on same line
		// $display("at %t have a cache line to update", $time);
        cache_mem_addr[found_index] <= wr_addr;
        cache_mem_data[found_index] <= cache_wr_data;
		cache_entry[found_index] <= 1'b1;
        cache_valid[wr_addr] <= 1'b1;
		// $display("L96 cache entry %t", $time);
      end : if_found_entry
      else  begin : else1_mt4wr// find an empty line
        for (int j=0; j<= CACHE_SIZE-1; j=j+1) begin : for1
          if(cache_entry[j]==0) begin : if2 // found empty line
            found_empty_line = 1'b1;
            found_index = j;
            success =1; 
            // $display("L100  at %t found line empty, found_index=%d", $time, found_index);
            break;
          end : if2
        end : for1
        if (found_empty_line) begin : update_cache
          cache_mem_addr[found_index] <= wr_addr;
          cache_mem_data[found_index] <= cache_wr_data;
          cache_entry[found_index] <= 1'b1;
          cache_valid[wr_addr]            <= 1'b1;  
          // $display("L109 cache entry %t", $time);
        end : update_cache
		// $display("found_empty_line = %b", found_empty_line);
        ap_no_cache_line_found_4write: assert(found_empty_line || found_existing_entry) else 
                $error("L111 attempt to write a cache line into non existant space %t", $time);
		ap_no_cache_line_found_4write_concurrent : assert property (@ (posedge clk) found_empty_line || found_existing_entry) else 
                $error("L111 attempt to write a cache line into non existant space %t", $time);
      end : else1_mt4wr
     end : wr_cache_updates
  end : cache_mem_handling


  // ********** READ from cache **********
  always @ (posedge clk) begin : cache_read
    automatic logic success =0;
    automatic int   i=0;
    if (check_cache && cache_exists) begin : if_exists 
      success=0;
      while (!success) begin : while1       
        if(cache_entry[i]==1 && cache_mem_addr[i] == addr) begin : if1 // found cache line 
          success =1;
          cache_data <=  cache_mem_data[i];
          //$display("found cache read");
        end : if1
        else i=i+1;
        ap_read_cache_line: assert (i!=CACHE_SIZE) else $error("error in reading cache line");
		ap_read_cache_line_concurrent : assert property (@ (posedge clk) i!=CACHE_SIZE) else $error("error in reading cache line");
        if(i==CACHE_SIZE) break;
      end : while1
	  // $display("success read = %b", success);
      assert (success) else $error("attempt to read from a cache but no line found ");
    end : if_exists
  end : cache_read
endmodule : cache_mem