import cache_ctl_pkg::*;
module cc_visual(
                        //user interface
                        input logic data_vld, busy, cache_valid, miss, 
                        input  logic rd, wr,
                        input logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr,                     
                        input logic [MEM_DATA_WIDTH-1:0]         rd_data, 
                        input logic  [MEM_DATA_WIDTH-1:0]         wr_data,  
                        // clock and resets 
                        input logic clk, 
                        input logic rst_n);

  // Local declarations
    timeunit 1ns;     timeprecision 100ps;
  // ***** TO PUT IN A checker  chk_rd_wr_cntlr 
   property p_wr_rd;
    int v_addr, v_data;
      first_match(($rose(wr), v_addr=addr, v_data=wr_data) ##1 
      !(wr && addr==v_addr)[*1:$] ##1 rd && addr==v_addr) |->
                                   ##[1:MEM_CYCLES] rd_data==v_data;
  endproperty : p_wr_rd
  ap_wr_rd : assert property(@ (posedge clk) p_wr_rd);

     mp_no_simultaneous_rd_wr: assume property (
     @ (posedge clk) not (rd && wr));

    mp_busy_no_rd_wr: assume property (
            @ (posedge clk) busy |-> !rd && !wr);
  // *****
  // ***** chk_rd_cntrl
  property p_read_miss;
	  logic [MAIN_MEM_ADDRESS_WIDTH-1:0] v_addr;
 	  (rd && miss, v_addr=addr) |-> ##MEM_CYCLES data_vld ##1 
        // cache_mem1.cache_valid[v_addr];
          cache_valid;
  endproperty : p_read_miss
  ap_read_miss : assert property(@ (posedge clk) p_read_miss);

  
endmodule :cc_visual 

class Cvar;
  rand logic data_vld, busy, cache_valid, miss;
  rand  logic rd, wr;
  rand logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr;
  rand logic [MEM_DATA_WIDTH-1:0]         rd_data;
  rand logic  [MEM_DATA_WIDTH-1:0]         wr_data;

  constraint c_addr {addr <10; };
  constraint c_wr_data {wr_data >10 && wr_data <20;};
  constraint crd_wr {
			   wr dist{
				 0 := 10,
				  1 := 3};
			   rd dist {
				 0 := 3,
     			 1 := 10};
					 };



endclass : Cvar

module cc_tb;
  logic data_vld, busy, cache_valid, miss;
  logic rd, wr;
  logic  [MAIN_MEM_ADDRESS_WIDTH-1:0] addr;
  logic [MEM_DATA_WIDTH-1:0]         rd_data;
  logic  [MEM_DATA_WIDTH-1:0]         wr_data;
  logic clk=1'b1, rst_n=1'b1;

  Cvar cvar_obj=new();

  cc_visual cc_visual1(.*);
  
  initial forever #5 clk=!clk;

  initial begin : test
    repeat (100) begin : r1
    @(posedge clk);	  
	  cvar_obj.randomize(); #1;
      data_vld <= cvar_obj.data_vld;
      busy <= cvar_obj.busy;
      cache_valid <= cvar_obj.cache_valid;
      miss <= cvar_obj.miss;
      rd <= cvar_obj.rd;
      wr <= cvar_obj.wr;
      addr <= cvar_obj.addr;
      rd_data <= cvar_obj.rd_data;
      wr_data <= cvar_obj.wr_data;
	  end : r1
    #1000 $finish;
  end  : test
  
  
endmodule : cc_tb





















