
// typedef logic [WIDTH-1 : 0] word_t;
import cache_ctl_pkg::*;
typedef logic [MAIN_MEM_ADDRESS_WIDTH-1 : 0] word_t;
typedef word_t [0 : (2**BIT_DEPTH-1)] buffer_t;

module fifo_mem ( input logic clk, rst_n, 
              output logic empty, full,
              output word_t data_out,
              output logic error,
              input word_t data_in,
              input push, pop);
  timeunit 1ns;  timeprecision 100ps;  
  // ==== from package:
  
  // =====
  //input clk; input reset_n;  output empty;  output almost_empty;
  // output almost_full;  output full;  output word_t data_out;
  // output error;  input word_t data_in; input push; input pop;
  // =====


  logic [BIT_DEPTH : 0] wr_addr; // write fifo address // BC 
  logic [BIT_DEPTH : 0] rd_addr; // read fifo address  // BC
  buffer_t buffer; // fifo storage
  parameter shiftVal = int'(2**BIT_DEPTH) ;

  // Push on full with no pop error 
  property p_push_error; 
    @ (posedge clk) 
      not (push && full && !pop); 
  endproperty : p_push_error
    ap_push_error_1 : assert property (p_push_error);

   property p_push_data; 
  word_t v_data;
  logic [BIT_DEPTH-1 : 0] v_ptr;
     (push, v_data = data_in, v_ptr = wr_addr) |=> buffer[v_ptr] == v_data; 
  endproperty : p_push_data
  ap_push_data : assert property (@ (posedge clk) p_push_data);
  
      // mp_push_error : assume property (p_push_error);
      
      // No pop on empty 
      property p_pop_error; 
        @ (posedge clk) 
          not (pop && empty); 
      endproperty : p_pop_error
        ap_pop_error_1 : assert property (p_pop_error);
          // mp_pop_error : assume property (p_pop_error);
          
          // RTL Detailed design
          logic [1:0]            push_pop;
          // Defining control for maintenance of FIFO word count
          assign             push_pop = {push , pop};    // temporary concatenation 
          // providing READ data to output
          assign data_out = buffer[rd_addr[BIT_DEPTH-1:0]];

          // purpose: writes data from FIFO and controls
          //          words within the FIFO.  It also provides the READ and
          //          WRITE Fifo Addresses. 
          
          always @ (posedge clk or negedge rst_n) 
          begin 
            if (!rst_n)  
            begin  // asynchronous reset (active low)
              rd_addr         <= 0;
              wr_addr         <= 0;
            end
            else
            begin
              case (push_pop)
                2'b00 :
                  ;              // no push, no pop
                
                2'b01 : 
                  begin                   // no push, pop for READ 
                    rd_addr         <= (rd_addr+1) ; // % shiftVal;
                  end   
                2'b10 :                    // push for WRITE, no pop
                  begin              
                    wr_addr         <= (wr_addr + 1) ; //% shiftVal;
                    buffer[wr_addr[BIT_DEPTH-1:0]] <= data_in;
                  end
                2'b11 :                    // push for WRITE, pop for READ
                  begin
                    rd_addr <= (rd_addr+1) ; //% shiftVal;
                    wr_addr <= (wr_addr + 1) ; //% shiftVal;
                    buffer[wr_addr[BIT_DEPTH-1:0]] <= data_in;
                  end
                default  :
                  begin
                    /*             a_illegal_fifo_cmd_1 : assert (1'b0) else
                     $warning ("%0t %0m Meta value detected in FIFO command, push_pop %2b ",
                     $time, push_pop);
                     */
		                a_illegal_fifo_cmd_1 : assert property (1'b0) ;
                end
                
              endcase
            end //else
          end //always

  // Reporting of flags
  // properties used as guide for this design since they define
  // the requirements
  
  assign  error = (pop && empty) || (push && full && !pop);
  assign full = (wr_addr[BIT_DEPTH] != rd_addr[BIT_DEPTH]); 
              //error : (wr_addr - rd_addr) == FULL;
  assign empty = (wr_addr == rd_addr);


endmodule : fifo_mem
